/* dhcpd.c
 *
 * docsis_server
 *
 * Large portions of this source code are
 * Copyright (C) 2002 Access Communications
 *                    email docsis_guy@accesscomm.ca
 *
 * Designed for use with DOCSIS cable modems and hosts.
 * And for use with MySQL Database Server.
 *
 ********************************************************************
 **  This server still slightly resembles:

 * Moreton Bay DHCP Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <pwd.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <net/if.h>
#include <linux/sockios.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>

#define DHCP_SERVER

#include "dhcpd.h"
#include "../src-mysql/my_sql.h"
#include "socket.h"
#include "options.h"
#include "../src-config/config_file.h"
#include "../Version.h"


// global variable
extern int dhcp2cfgen[2];

// prototypes
void send_positive_message( dhcp_message *message, int mess_type );
void            send_Offer( dhcp_message *message );
void              send_ACK( dhcp_message *message );
void              send_NAK( dhcp_message *message );
void        send_Query_ACK( dhcp_message *message );
void        send_Query_NAK( dhcp_message *message );
void            leaseQuery( dhcp_message *message );


void dhcp_server( void ) {
	char			*iface_pref, *euid_pref, *dbhost_pref, *force_chroot;
	dhcp_message		message;
	u_int32_t		server_ip;
	struct passwd		*pass_struct;

	InitSQLconnection();
	my_LoadChecker();

	iface_pref = my_GetDHCPint();
	server_ip = getInterfaceIP( iface_pref );
	init_DHCP_Socket( server_ip );

	dbhost_pref = GetConfigVar( "mysql-host" );
	euid_pref = GetConfigVar( "effective-userid" );
	force_chroot = GetConfigVar( "force-localhost-chroot" );
	pass_struct = getpwnam( euid_pref );

	if (strcmp(dbhost_pref,"localhost") || !strcmp(force_chroot, "yes")) {
		// We cannot chroot if connect to the DB via
		// a unix socket vs tcp
		chroot( CHROOT_PATH );
	}
	if (pass_struct) {
		seteuid( pass_struct->pw_uid );
		setegid( pass_struct->pw_gid );
	}

	my_syslog( MLOG_DHCP, "docsis_server DHCP version %s activated", VERSION);

	while (1) { /* loop until universe collapses */

		/* Clear Message Struct */
		bzero( &message, sizeof( dhcp_message ) );

		/* Get Packet. If error start another while loop */
		if (getPacket( &message ) < 0) continue;
		message.server_ip = server_ip;

 		DecodeOptions( &message );
		if (message.in_opts.message_type == 0) {
			message.in_opts.message_type = DHCP_REQUEST;
		}

		switch( message.in_opts.message_type ) {
		case DHCP_DISCOVER:
			send_Offer( &message );
			break;
			
		case DHCP_REQUEST:
			send_ACK( &message );
			break;

		case DHCP_RELEASE:
			/* checkRelease( &message ); */
			break;
		case DHCP_INFORM:
			send_ACK( &message );
			break;

		case DHCP_LEASE_QUERY:	/* wierd ubR thingy  0x0d */
			leaseQuery( &message );
			break;
			
		default: {
			my_syslog(LOG_WARNING, "unsupported DHCP message (%02x) %s -- ignoring",
				message.in_opts.message_type, message.s_macaddr );
			}
		}
	}

	/* should never executes */
	my_syslog(LOG_ERR, "exit");
	closelog();
	exit(0);
}



void send_positive_message( dhcp_message *message, int mess_type ) {
	static char	*newip_pref = NULL;
	u_int32_t	lease_time;

	/* check SQL table for this MAC address */
	my_findMAC( message );

	if (message->lease_type == LEASE_NOT_FOUND && message->in_opts.docsis_modem == 1 ) {
		my_syslog(LOG_INFO, "NAK -- unknown docsis modem %s", message->s_macaddr );
		send_NAK( message );
		return;
	}

	if (message->lease_type == LEASE_REJECT) { return; }

	if (message->lease_type == LEASE_CPE &&
	    message->cpe_type == CPE_DYNAMIC &&
	    message->lockip == 0 &&
	    message->lease_time < 60 ) {
		// Your Lease has expired. You get a new IP!
		my_DeleteLease( message );
		message->lease_type = LEASE_NOT_FOUND;
	}

	if (message->lease_type == LEASE_CPE ) {
		if ( Verify_Vlan( message ) ) {
			my_syslog(LOG_WARNING, "NEW -- GIaddr does not match vlan - gi %s ip %s vlan %d",
				message->in_giaddr, message->s_ipaddr, message->vlan );
			if (message->cpe_type == CPE_STATIC) {
				my_DeleteLease( message );
				message->lease_type = LEASE_NOT_FOUND;
			} else {
				// if message->cpe_type == CPE_DYNAMIC
				message->lease_type = LEASE_NOAUTH;
			}
		}
	}

	if (message->lease_type == LEASE_NOT_FOUND && message->in_opts.agent_mac_len == 0) {
		my_syslog(LOG_WARNING, "NAK -- CPE asking for new IP directly - mac %s",
			message->s_macaddr );
		send_NAK( message );
		return;
	}

	if (newip_pref == NULL) newip_pref = GetConfigVar( "assign-new-ip" );

	if (message->lease_type == LEASE_NOT_FOUND) {
		if (*newip_pref == 'n') return;

		my_getNewIP( message );

		if (message->ipaddr == 0) {
			send_NAK( message );
			return;
		}
		if (message->lease_type == LEASE_CPE) {
			if (message->cpe_type == CPE_STATIC) {
				my_syslog(LOG_INFO, "NEW STATIC -- mac %s ip %s vlan %d",
					message->s_macaddr, message->s_ipaddr, message->vlan );
			}
			if (message->cpe_type == CPE_DYNAMIC) {
				my_syslog(LOG_INFO, "NEW DYNAMIC -- mac %s ip %s vlan %d",
					message->s_macaddr, message->s_ipaddr, message->vlan );
			}
		}
		if (message->lease_type == LEASE_NOAUTH) {
			my_syslog(LOG_INFO, "NEW NOAUTH -- mac %s ip %s vlan %d",
				message->s_macaddr, message->s_ipaddr, message->vlan );
		}
	}

	if ( my_Get_Net( message ) ) {
		my_syslog(LOG_WARNING, "NAK -- no subnet for ip %s mac %s vlan %d",
			message->s_ipaddr, message->s_macaddr, message->vlan );
		send_NAK( message );
		return;
	}

	if (message->in_opts.request_addr != 0 ) {
		if (message->ipaddr != message->in_opts.request_addr ) {
			my_syslog(LOG_ERR, "WARN -- bad ip requested %s mac %s real-ip %s",
				message->in_opts.s_request_addr, message->s_macaddr, message->s_ipaddr );
		}
	}

	message->out_pack.op    = BOOTREPLY;
	message->out_pack.htype = ETH_10MB;
	message->out_pack.hlen  = ETH_10MB_LEN;
	message->out_pack.hops  = message->in_pack.hops;
	message->out_pack.flags = message->in_pack.flags;
	message->out_pack.xid   = message->in_pack.xid;

	message->out_pack.ciaddr = message->in_pack.ciaddr;
	message->out_pack.yiaddr = message->ipaddr;
	message->out_pack.siaddr = message->server_ip;
	message->out_pack.giaddr = message->in_pack.giaddr;
	message->out_pack.cookie = DHCP_MAGIC;

	memcpy( message->out_pack.macaddr, message->macaddr, 6 );

	initOpt( message );

	if (mess_type == DHCP_OFFER) {
		addOpt( 0x35, 0x01, "\x02");
	}
	if (mess_type == DHCP_ACK) {
		addOpt( 0x35, 0x01, "\x05");
	}

	// subnet mask
	addOpt( 0x01, 0x04, (char *) &(message->netptr->mask) );

	// gateway
	addOpt( 0x03, 0x04, (char *) &(message->netptr->gateway) );

	// Broadcast Address
	addOpt( 0x1c, 0x04, (char *) &(message->netptr->broadcast) );

	// Lease Times
	if (message->cpe_type == CPE_STATIC || message->lockip == 1) {
		message->lease_time = message->netptr->lease_time;
		lease_time = htonl( message->lease_time );
		addOpt( 0x33, 0x04, (char *) &lease_time );
		lease_time = htonl( message->lease_time - 300 );
		addOpt( 0x3A, 0x04, (char *) &lease_time );
		addOpt( 0x3B, 0x04, (char *) &lease_time );
	} else {
		// if message->cpe_type == CPE_DYNAMIC
		lease_time = htonl( (u_int32_t) message->lease_time );
		if (message->lease_time < 0) lease_time = htonl( message->netptr->lease_time / 2 );
		addOpt( 0x33, 0x04, (char *) &lease_time );
		lease_time = ntohl( lease_time ) - 15;
		if (lease_time < 0) lease_time = 5;
		lease_time = htonl( lease_time );
		addOpt( 0x3A, 0x04, (char *) &lease_time );
		addOpt( 0x3B, 0x04, (char *) &lease_time );
	}

	addBigOpt( my_Get_Opt( message->netptr->opt1 ) );
	addBigOpt( my_Get_Opt( message->netptr->opt2 ) );
	addBigOpt( my_Get_Opt( message->netptr->opt3 ) );

	/* Config File Name */
	if (message->lease_type == LEASE_CM) {
		int	cflen = 127;
		if (cflen > MAX_CONFIG_FILE_NAME) { cflen = MAX_CONFIG_FILE_NAME; }
		strncpy( message->out_pack.file, message->cfname, cflen );
		addOpt( 0x43, strlen( message->out_pack.file ), message->cfname );
	}

	/* Add Terminator */
	addOpt( 0xff, 0x00, NULL );

	my_dhcplog( mess_type, message );

	sendPacket( message );

	if (mess_type == DHCP_OFFER && message->lease_type == LEASE_CM) {
		// If a CM is booting up send message to the config-file generator server
		// write( dhcp2cfgen[1], &(message->old_ipaddr), 4 );
		// write( dhcp2cfgen[1], &(message->ipaddr), 4 );

		// Update AgentID of cablemodem
		my_UpdateAgent( message );
	}

	/* if a customer pc, update the leases table, if a cable modem dont bother */
	if (message->lease_type == LEASE_CPE || message->lease_type == LEASE_NOAUTH ) {
		my_UpdateLease( message );
	}
}

/* send a DHCP OFFER to a DHCP DISCOVER */
void send_Offer( dhcp_message *message ) {
	send_positive_message( message, DHCP_OFFER );
}

/* send a DHCP ACK to a DHCP REQUEST */
void send_ACK( dhcp_message *message ) {
	send_positive_message( message, DHCP_ACK );
}


void send_NAK( dhcp_message *message ) {
	message->out_pack.op    = BOOTREPLY;
	message->out_pack.htype = ETH_10MB;
	message->out_pack.hlen  = ETH_10MB_LEN;
	message->out_pack.hops  = message->in_pack.hops;
	message->out_pack.flags = message->in_pack.flags;
	message->out_pack.xid   = message->in_pack.xid;

	message->out_pack.ciaddr = message->in_pack.ciaddr;
	message->out_pack.siaddr = message->server_ip;
	message->out_pack.giaddr = message->in_pack.giaddr;
	message->out_pack.cookie = DHCP_MAGIC;

	memcpy( message->out_pack.macaddr, message->macaddr, 6 );

	initOpt( message );
	addOpt( 0x35, 0x01, "\x06" );  /* DHCP NAK */
	addOpt( 0x36, 0x04, (char *) &message->server_ip );
	addOpt( 0xff, 0x00, NULL );

	my_dhcplog( DHCP_NAK, message );

	sendPacket( message );
}

void send_Query_NAK( dhcp_message *message ) {
	send_NAK( message );
}

void send_Query_ACK( dhcp_message *message ) {
	u_int32_t	lease_time;

	message->out_pack.op    = BOOTREPLY;
	message->out_pack.htype = ETH_10MB;
	message->out_pack.hlen  = ETH_10MB_LEN;
	message->out_pack.hops  = message->in_pack.hops;
	message->out_pack.flags = message->in_pack.flags;
	message->out_pack.xid   = message->in_pack.xid;

	message->out_pack.ciaddr = message->in_pack.ciaddr;
	message->out_pack.yiaddr = message->ipaddr;
	message->out_pack.siaddr = message->server_ip;
	message->out_pack.giaddr = message->in_pack.giaddr;
	message->out_pack.cookie = DHCP_MAGIC;

	memcpy( message->out_pack.macaddr, message->macaddr, 6 );

	if (message->cpe_type == CPE_DYNAMIC) lease_time = htonl( (u_int32_t) message->lease_time );
	if (message->cpe_type == CPE_STATIC) lease_time = htonl( 3600 * 12 );  // give em 12 hours

	initOpt( message );
	addOpt( 0x35, 0x01, "\x05" );
	addOpt( 0x33, 0x04, (char *) &lease_time );
	lease_time = ntohl( lease_time ) - 60;
	if (lease_time < 0) lease_time = 5;
	lease_time = htonl( lease_time );
	addOpt( 0x3A, 0x04, (char *) &lease_time );
	addOpt( 0x3B, 0x04, (char *) &lease_time );
	if( my_findAgent( message ) ) {
		addOpt( 0x52, message->in_opts.agent_mac_len, NULL );
		addOpt( 0x01, message->in_opts.agentid_len,   message->in_opts.agentid );
		addOpt( 0x02, message->in_opts.modem_mac_len, message->in_opts.modem_mac );
	}
	addOpt( 0xff, 0x00, NULL );

	my_dhcplog( DHCP_QUERY_ACK, message );

	my_UpdateTime( message );

	sendPacket( message );
}



void	leaseQuery( dhcp_message *message ) {
	if ( (message->in_pack.htype != 0) &&
	     (message->in_pack.hlen  != 0) ) {
		/* Search by MAC address */
		my_findMAC( message );

		if (message->lease_type == LEASE_REJECT) return;

		if (message->lease_type == LEASE_CPE &&
		    message->cpe_type == CPE_DYNAMIC &&
		    message->lockip == 0 &&
		    message->lease_time < 0 ) {
			// Your Lease has expired. You get a new IP!
			my_DeleteLease( message );
			message->lease_type = LEASE_NOT_FOUND;
		}

		if (message->lease_type == LEASE_NOT_FOUND ) {
			send_Query_NAK( message );
		} else {
			send_Query_ACK( message );
		}
		return;
	} else if (message->in_pack.ciaddr != 0) {
		/* Search by IP -- The majority of requests from Cisco CMTS*/
		my_findIP( message, FIND_CIADDR );

		if (message->lease_type == LEASE_REJECT) return;

		if (message->lease_type == LEASE_CPE &&
		    message->cpe_type == CPE_DYNAMIC &&
		    message->lockip == 0 &&
		    message->lease_time < 0 ) {
			// Your Lease has expired. You get a new IP!
			my_DeleteLease( message );
			message->lease_type = LEASE_NOT_FOUND;
		}

		if (message->lease_type == LEASE_NOT_FOUND ) {
			send_Query_NAK( message );
		} else {
			send_Query_ACK( message );
		}
		return;
	} else if (message->in_opts.client_ident_len > 0) {
		/* Search by dhcp-client-ident flag */
		/* Im not sure how I will do this. Log it */
		char	tmpstr[4096];
		cv_macaddr( message->in_opts.client_ident, tmpstr,
			    message->in_opts.client_ident_len );
		my_syslog( LOG_INFO, "DHCPlq - dhcp-client-ident search -%s-", tmpstr );
	}
}
